--***********************************************************
--**                    THE INDIE STONE                    **
--**				  Author: turbotutone				   **
--**				                        			   **
--**				  Modified by: RedChili				   **
--**			To use with mod: BetterBuilding     	   **
--***********************************************************

require "ISUI/ISCollapsableWindow"

ISCustomBuildWindow = ISCollapsableWindow:derive("ISBuildWindow");

ISCustomBuildWindow.savedX = 100
ISCustomBuildWindow.savedY = 100
ISCustomBuildWindow.savedWidth = 600
ISCustomBuildWindow.savedHeight = 400

BuildingAssignmentMode = false
BuildingAssignmentRecipe = nil

function ISCustomBuildWindow:saveConfig()
    local player = getSpecificPlayer(0)
    local modData = player:getModData()
    modData.ISBuildWindowConfig = {
        x = self:getX(),
        y = self:getY(),
        width = self:getWidth(),
        height = self:getHeight()
    }
end

function ISCustomBuildWindow:loadConfig()
    local player = getSpecificPlayer(0)
    local modData = player:getModData()
    if modData.ISBuildWindowConfig then
        ISCustomBuildWindow.savedX = modData.ISBuildWindowConfig.x or ISCustomBuildWindow.savedX
        ISCustomBuildWindow.savedY = modData.ISBuildWindowConfig.y or ISCustomBuildWindow.savedY
        ISCustomBuildWindow.savedWidth = modData.ISBuildWindowConfig.width or ISCustomBuildWindow.savedWidth
        ISCustomBuildWindow.savedHeight = modData.ISBuildWindowConfig.height or ISCustomBuildWindow.savedHeight
    end
end

function ISCustomBuildWindow:initialise()
    ISCollapsableWindow.initialise(self);

    if self.maximumHeightPercent>0 then
        local maxPercent = math.min(self.maximumHeightPercent, 1);
        self.maximumHeight = getCore():getScreenHeight() * maxPercent;
    end
end

function ISCustomBuildWindow:createChildren()
    ISCollapsableWindow.createChildren(self);

    self.windowHeader = ISXuiSkin.build(self.xuiSkin, nil, ISBuildWindowHeader, 0, 0, 10, 10, self.player);
    if self.isoObject then
		local header = getText("IGUI_BuildingWindow_Header");
		local props = self.isoObject:getProperties();
		local surface = (props and props:Is("IsMoveAble") and props:Is("CustomName") and props:Val("CustomName")) or getText("IGUI_BuildingWindow_Surface");
        self.windowHeader.titleStr = header .. surface;
    end
    self.windowHeader:initialise();
    self.windowHeader:instantiate();
    self:addChild(self.windowHeader);

    self.BuildPanel = ISXuiSkin.build(self.xuiSkin, nil, ISBuildPanel, 0, 0, 10, 10, self.player, nil, self.isoObject);

    self.BuildPanel.recipeQuery = "AnySurfaceCraft";

    self.BuildPanel:initialise();
    self.BuildPanel:instantiate();
    self:addChild(self.BuildPanel);

    if self.pinButton then
        self.pinButton:setAnchorRight(false);
    end
    if self.collapseButton then
        self.collapseButton:setAnchorRight(false);
    end

    self.resizeWidget.resizeFunction = ISCustomBuildWindow.calculateLayout;
    self.resizeWidget2.resizeFunction = ISCustomBuildWindow.calculateLayout;

    self:xuiRecalculateLayout();
end

function ISCustomBuildWindow:xuiRecalculateLayout(_preferredWidth, _preferredHeight, _force, _anchorRight)
    if self.calculateLayout and ((not self.dirtyLayout) or _force) then
        self.xuiPreferredResizeWidth = self.width;
        self.xuiPreferredResizeHeight = self.height;
        self.xuiResizeAnchorRight = _anchorRight;
        if _preferredWidth then
            self.xuiPreferredResizeWidth = _preferredWidth<0 and self.width+_preferredWidth or _preferredWidth;
        end
        if _preferredHeight then
            self.xuiPreferredResizeHeight = _preferredHeight<0 and self.height+_preferredHeight or _preferredHeight;
        end
        self.dirtyLayout = true;
    end
end

function ISCustomBuildWindow:calculateLayout(_preferredWidth, _preferredHeight)
    self:validateSizeBounds();

    local th = self:titleBarHeight();
    local rh = self.resizable and self:resizeWidgetHeight() or 0;

    local width = math.max(_preferredWidth or 0, self.minimumWidth);
    local height = math.max(_preferredHeight or 0, self.minimumHeight);

    if self.maximumWidth>0 then width = math.min(width, self.maximumWidth); end
    if self.maximumHeight>0 then height = math.min(height, self.maximumHeight); end

    self.windowHeader:calculateLayout(0, 0);
    self.BuildPanel:calculateLayout(0, 0);

    width = math.max(width, self.windowHeader:getWidth());
    width = math.max(width, self.BuildPanel:getWidth());


    local x,y = 0,th;

    self.windowHeader:setX(0);
    self.windowHeader:setY(th);
    self.windowHeader:calculateLayout(width, 0);

    local wh = self.windowHeader:getHeight();

    if self.BuildPanel then
        self.BuildPanel:setX(0);
        self.BuildPanel:setY(th+wh);
        self.BuildPanel:calculateLayout(width, math.max(0, height-th-rh-wh));

        width = math.max(width, self.BuildPanel:getWidth());
        height = math.max(height, self.BuildPanel:getHeight()+th+rh+wh);
    end


    self:setWidth(width);
    self:setHeight(height);

    self.dirtyLayout = false;

    if self.pinButton then
        self.pinButton:setX(width - 3 - self.pinButton:getWidth())
    end
    if self.collapseButton then
        self.collapseButton:setX(width - 3 - self.collapseButton:getWidth())
    end
end

function ISCustomBuildWindow:prerender()
    self:stayOnSplitScreen();

    if self.dirtyLayout then
        local oldX = self:getX();
        local oldWidth = self:getWidth();
        if self.calculateLayout then self:calculateLayout(self.xuiPreferredResizeWidth, self.xuiPreferredResizeHeight); end
        self.dirtyLayout = false;

        if self.xuiResizeAnchorRight then
            self:setX(oldX - (self:getWidth()-oldWidth))
            self.xuiResizeAnchorRight = false;
        end
    end

    ISCollapsableWindow.prerender(self);
end

function ISCustomBuildWindow:stayOnSplitScreen()
    ISUIElement.stayOnSplitScreen(self, self.playerNum)
end


function ISCustomBuildWindow:render()
    self:stayOnSplitScreen();
    ISCollapsableWindow.render(self);
    self:renderJoypadNavigateOverlay(self.playerNum)
end

function ISCustomBuildWindow:update()
    ISCollapsableWindow.update(self);

    local valid = false;

    if self.isoObject and self.player then
        local dist = self.panelCloseDistance or 10;
        local ex, ey = self.isoObject:getX(), self.isoObject:getY();
        local px, py = self.player:getX(), self.player:getY();

        if self.isoObject:getSquare() and px > ex-dist and px < ex+dist and py > ey-dist and py < ey+dist then
            valid = true;
        end
    else
        valid = self.player;
    end
    if self.hasClosedWindowInstance then
        return;
    end

    if not valid then
       self:close();
    end
    return valid;
end

function ISCustomBuildWindow:close()
    if self.hasClosedWindowInstance then return end
    self.hasClosedWindowInstance = true

    if self.BuildPanel then
        self.BuildPanel:close()
    end

    self:saveConfig() -- SAVE CONFIG HERE

    ISCollapsableWindow.close(self)
    ISEntityUI.OnCloseWindow(self)
    if JoypadState.players[self.playerNum+1] then
        if self.unfocusRecursive then
            self:unfocusRecursive(getFocusForPlayer(self.playerNum), self.playerNum)
        elseif isJoypadFocusOnElementOrDescendant(self.playerNum, self) then
            setJoypadFocus(self.playerNum, nil)
        end
    end
    if self.entity then
        self.entity:setUsingPlayer(nil)
    end
    if self.isoObject and getCore():getOptionDoContainerOutline() then
        self.isoObject:setOutlineHighlight(false)
        self.isoObject:setOutlineHlAttached(false)
    end
    self:removeFromUIManager()
    ISCustomBuildWindow.instance = nil -- RMOVE CUSTOM INSTANCE HERE
end



function ISCustomBuildWindow:validateSizeBounds()
    if (not self.minimumWidth) or self.minimumWidth<0 then self.minimumWidth = 0; end
    if (not self.minimumHeight) or self.minimumHeight<0 then self.minimumHeight = 0; end

    if (not self.maximumWidth) or self.maximumWidth<0 then self.maximumWidth = 0; end
    if (not self.maximumHeight) or self.maximumHeight<0 then self.maximumHeight = 0; end

    if self.maximumWidth>0 and self.maximumWidth<self.minimumWidth then
        self.maximumWidth = self.minimumWidth;
    end
    if self.maximumHeight>0 and self.maximumHeight<self.minimumHeight then
        self.maximumHeight = self.minimumHeight;
    end
end

function ISCustomBuildWindow:onGainJoypadFocus(joypadData)
    ISCollapsableWindow.onGainJoypadFocus(self, joypadData)
    local recipeCategories = self.BuildPanel.recipeCategories.recipeCategoryPanel
    recipeCategories:setJoypadFocused(true, joypadData)
end

function ISCustomBuildWindow:onJoypadDown_Descendant(descendant, button, joypadData)
    if button == Joypad.AButton then
        local craftControl = self.BuildPanel.craftRecipePanel.craftControl
        craftControl.buttonCraft:forceClick()
    end
    if button == Joypad.BButton then
        self:close()
    end
end

function ISCustomBuildWindow:onJoypadNavigateStart_Descendant(descendant, joypadData)
    local recipeCategories = self.BuildPanel.recipeCategories.recipeCategoryPanel
    local recipeFilterPanel = self.BuildPanel.recipesPanel.recipeFilterPanel
    local recipeIconPanel = self.BuildPanel.recipesPanel.recipeIconPanel
    local recipeListPanel = self.BuildPanel.recipesPanel.recipeListPanel.recipeListPanel
    local listOrIconPanel = recipeIconPanel:isVisible() and recipeIconPanel or recipeListPanel
    local recipePanel = self.BuildPanel.craftRecipePanel
    local recipeInputs = recipePanel and recipePanel.inputs or nil
    local craftControl = recipePanel and recipePanel.craftControl or nil
    local inventoryPanel = nil
    recipeCategories.joypadNavigate = { right = listOrIconPanel }
    recipeFilterPanel.joypadNavigate = { left = recipeCategories, right = recipeInputs, down = listOrIconPanel }
    listOrIconPanel.joypadNavigate = { left = recipeCategories,  up = recipeFilterPanel, right = recipeInputs }
    if recipeInputs then
        recipeInputs.joypadNavigate = { left = listOrIconPanel, right = inventoryPanel, down = craftControl }
    end
    if craftControl then
        craftControl.joypadNavigate = { left = listOrIconPanel, right = inventoryPanel, up = recipeInputs }
    end
end

function ISCustomBuildWindow:new(x, y, width, height, player, isoObject, queryOverride)
    local o = ISCollapsableWindow:new(x, y, width, height);
    setmetatable(o, self)
    self.__index = self

    o.x = x;
    o.y = y;
    o.player = player;
    o.playerNum = player:getPlayerNum();
    o.borderColor = {r=0.4, g=0.4, b=0.4, a=1};
    o.backgroundColor = {r=0, g=0, b=0, a=0.8};
    o.width = width;
    o.height = height;
    o.anchorLeft = true;
    o.anchorRight = false;
    o.anchorTop = true;
    o.anchorBottom = false;
    o.pin = true;
    o.isCollapsed = false;
    o.collapseCounter = 0;
    o.title = nil;
    o.drawFrame = true;

    o.panelCloseDistance = 8;
    o.isoObject = isoObject;
    o.queryOverride = queryOverride;

    o.resizable = true;
    o.enableHeader = true;

    o.minimumWidth = 600;
    o.minimumHeight = 400;

    o.maximumWidth = 0;
    o.maximumHeight = 0;

    o.maximumHeightPercent = -1;

    o.overrideBPrompt = true;
    o:setWantKeyEvents(true)

    print("Build window meta table: " ..tostring(getmetatable(self)))

    if isoObject and getCore():getOptionDoContainerOutline() then
        isoObject:setOutlineHighlight(true);
        isoObject:setOutlineHlAttached(true);
        isoObject:setOutlineHighlightCol(1, 1, 1, 1);
    end

    return o
end

ISCustomBuildWindow.openNew = function()
    ISCustomBuildWindow:loadConfig();
    local x = ISCustomBuildWindow.savedX or 100;
    local y = ISCustomBuildWindow.savedY or 100;
    local width = ISCustomBuildWindow.savedWidth or 600;
    local height = ISCustomBuildWindow.savedHeight or 400;
    ISCustomBuildWindow.instance = ISCustomBuildWindow:new(x, y, width, height, getSpecificPlayer(0), nil, nil);
    ISCustomBuildWindow.instance:initialise();
    ISCustomBuildWindow.instance:addToUIManager();
end

ISCustomBuildWindow.onPressKey = function(key)
    if BuildingAssignmentMode then
        local slot = -1;
        if getCore():isKey("Hotbar 1", key) then slot = 1; end
        if getCore():isKey("Hotbar 2", key) then slot = 2; end
        if getCore():isKey("Hotbar 3", key) then slot = 3; end
        if getCore():isKey("Hotbar 4", key) then slot = 4; end
        if getCore():isKey("Hotbar 5", key) then slot = 5; end
        if getCore():isKey("Hotbar 6", key) then slot = 6; end
        if getCore():isKey("Hotbar 7", key) then slot = 7; end
        if getCore():isKey("Hotbar 8", key) then slot = 8; end
        if slot ~= -1 then
            if ISBuildingHotbar.instance then
                print(BuildingAssignmentRecipe)
                ISBuildingHotbar.instance:assignRecipeToSlot(slot, BuildingAssignmentRecipe);
                print("Recipe assigned to hotbar slot " .. slot);
                ISBuildingHotbar.instance.assignmentPrompt:removeFromUIManager()
                ISBuildingHotbar.instance.assignmentPrompt = nil
            else
                print("No building hotbar instance found.");
            end
            BuildingAssignmentMode = false;
            BuildingAssignmentRecipe = nil;
            return;
        end
    end

    if not MainScreen.instance or not MainScreen.instance.inGame or MainScreen.instance:getIsVisible() then
        return;
    end

    if getCore():isKey("Building UI", key) and (Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) or Keyboard.isKeyDown(Keyboard.KEY_RSHIFT)) then
        if ISCustomBuildWindow.instance and ISCustomBuildWindow.instance:getIsVisible() then
            ISCustomBuildWindow.instance:close();
        else
            ISCustomBuildWindow.openNew();
        end
    end
end

local numberKeyThreshold = 150
local numberKeyPressTime = {}

local originalHotbarKeyStartPressed = ISHotbar.onKeyStartPressed or function(key) end
local originalHotbarKeyKeepPressed = ISHotbar.onKeyKeepPressed
local originalHotbarOnKeyPressed = ISHotbar.onKeyPressed

ISHotbar.onKeyStartPressed = function(key)
    if ISCustomBuildWindow.instance and 
       (key == Keyboard.KEY_1 or key == Keyboard.KEY_2 or key == Keyboard.KEY_3 or 
        key == Keyboard.KEY_4 or key == Keyboard.KEY_5 or key == Keyboard.KEY_6 or 
        key == Keyboard.KEY_7 or key == Keyboard.KEY_8) then
        numberKeyPressTime[key] = getTimestampMs()
        return true
    end
    return originalHotbarKeyStartPressed(key)
end

ISHotbar.onKeyKeepPressed = function(key)
    if ISCustomBuildWindow.instance and 
       (key == Keyboard.KEY_1 or key == Keyboard.KEY_2 or key == Keyboard.KEY_3 or 
        key == Keyboard.KEY_4 or key == Keyboard.KEY_5 or key == Keyboard.KEY_6 or 
        key == Keyboard.KEY_7 or key == Keyboard.KEY_8) then
        return true
    end
    return originalHotbarKeyKeepPressed(key)
end

ISHotbar.onKeyPressed = function(key)
    if ISCustomBuildWindow.instance and 
       (key == Keyboard.KEY_1 or key == Keyboard.KEY_2 or key == Keyboard.KEY_3 or 
        key == Keyboard.KEY_4 or key == Keyboard.KEY_5 or key == Keyboard.KEY_6 or 
        key == Keyboard.KEY_7 or key == Keyboard.KEY_8) then

        local pressTime = numberKeyPressTime[key] or 0
        local elapsed = getTimestampMs() - pressTime
        if elapsed < numberKeyThreshold then
            return originalHotbarOnKeyPressed(key)
        else
            return true
        end
    end
    return originalHotbarOnKeyPressed(key)
end

function AssignRecipeOnMouseDown(_self, _recipe)
    print("Custom hook triggered for recipe: " .. tostring(_recipe:getName()))
    local slot = -1
    if Keyboard.isKeyDown(Keyboard.KEY_1) then slot = 1 end
    if Keyboard.isKeyDown(Keyboard.KEY_2) then slot = 2 end
    if Keyboard.isKeyDown(Keyboard.KEY_3) then slot = 3 end
    if Keyboard.isKeyDown(Keyboard.KEY_4) then slot = 4 end
    if Keyboard.isKeyDown(Keyboard.KEY_5) then slot = 5 end
    if Keyboard.isKeyDown(Keyboard.KEY_6) then slot = 6 end
    if Keyboard.isKeyDown(Keyboard.KEY_7) then slot = 7 end
    if Keyboard.isKeyDown(Keyboard.KEY_8) then slot = 8 end

    if slot ~= -1 and ISBuildingHotbar.instance then
         ISBuildingHotbar.instance:assignRecipeToSlot(slot, _recipe)
         print("Assigned recipe " .. tostring(_recipe:getName()) .. " to slot " .. slot)
    end
end

local originalRecipeListInitialise = ISWidgetRecipeListPanel.createChildren

function ISWidgetRecipeListPanel:createChildren()
    originalRecipeListInitialise(self)
    self.recipeListPanel:setOnMouseDownFunction(self, function(_self, _recipe) _self.logic:setRecipe(_recipe) AssignRecipeOnMouseDown(_self, _recipe) end);
end

Events.OnCustomUIKey.Add(ISCustomBuildWindow.onPressKey)