require "ISUI/ISRadialMenu"

ISCustomRadialMenu = ISBaseObject:derive("ISCustomRadialMenu")

function ISCustomRadialMenu:init()
	ISCustomRadialMenu.defaultMenu = {};
	ISCustomRadialMenu.menu = ISCustomRadialMenu.defaultMenu;
end

function ISCustomRadialMenu:new(character, mode)
	local o = ISBaseObject.new(self)
	o.character = character
	o.playerNum = character:getPlayerNum()
    o.lastFocus = nil
    o.mode = mode
	ISUIEmoteConfig:readFile();
    ISCustomRadialMenu.instance = o
	return o
end

function ISCustomRadialMenu:display()
	local menu = getPlayerRadialMenu(self.playerNum)
	self:center()
	menu:addToUIManager()
	if JoypadState.players[self.playerNum+1] then
        local gamepadData = JoypadState.joypads[self.playerNum+1]
        self.lastFocus = gamepadData.lastfocus
		setJoypadFocus(self.playerNum, menu)
		self.character:setJoypadIgnoreAimUntilCentered(true)
	end
end

local originalPadDown = ISRadialMenu.onJoypadDown

function ISRadialMenu:onJoypadDown(button, joypadData)
	if button == Joypad.BButton or button == Joypad.XButton and ISCustomRadialMenu.instance then
		self:undisplay()
        setJoypadFocus(0, ISCustomRadialMenu.instance.lastFocus)
        ISCustomRadialMenu.instance = nil
        return
	end
    originalPadDown(self, button, joypadData)
end

local originalTriggerRelease = ISRadialMenu.onJoypadButtonReleased

function ISRadialMenu:onJoypadButtonReleased(button, joypadData)
    if not ISCustomRadialMenu.instance then
        originalTriggerRelease(self, button, joypadData)
        return
    end
    if tostring(button) == "5" and ISCustomRadialMenu.instance.mode == "assign" then
        local menu = getPlayerRadialMenu(self.playerNum)
        local sliceIndex = self.javaObject:getSliceIndexFromJoypad(self.joyfocus.id)
        local slot = sliceIndex + 1
        if ISBuildingHotbar.instance then
            ISBuildingHotbar.instance:assignRecipeToSlot(slot, ISBuildingHotbar.instance.selectedRecipe)
            print("Assigned recipe " .. tostring(ISBuildingHotbar.instance.selectedRecipe:getName()) .. " to slot " .. slot)
            if ISCustomRadialMenu.instance then
                menu:setSliceText(slot, ISBuildingHotbar.instance.selectedRecipe:getName())
                menu:setSliceTexture(slot, ISBuildingHotbar.instance.selectedRecipe:getIconTexture())
            end
        else
            print("Assignment failed: no recipe or hotbar instance found.")
        end
    elseif tostring(button) == "100" and ISCustomRadialMenu.instance.mode == "build" then
        local sliceIndex = self.javaObject:getSliceIndexFromJoypad(self.joyfocus.id)
        local slot = sliceIndex + 1
        if ISBuildingHotbar.instance then
            ISBuildingHotbar.instance:buildRecipe(slot, false)
            self:undisplay()
            ISCustomRadialMenu.instance = nil
            RBButtonHeld = false
            return
        end
    else
        originalTriggerRelease(self, button, joypadData)
    end
end


function ISCustomRadialMenu:center()
	local menu = getPlayerRadialMenu(self.playerNum)

	local x = getPlayerScreenLeft(self.playerNum)
	local y = getPlayerScreenTop(self.playerNum)
	local w = getPlayerScreenWidth(self.playerNum)
	local h = getPlayerScreenHeight(self.playerNum)

	x = x + w / 2
	y = y + h / 2

	menu:setX(x - menu:getWidth() / 2)
	menu:setY(y - menu:getHeight() / 2)
end

function ISCustomRadialMenu:fillMenu(submenu)
    local menu = getPlayerRadialMenu(self.playerNum)
    menu:clear()
    local hotbar = ISBuildingHotbar.instance
    for i = 1, 8 do
        local sliceText = "Slot " .. i
        local icon = nil
        if hotbar and hotbar.attachedRecipes[i] then
            local recipe = hotbar.attachedRecipes[i]
            icon = recipe:getIconTexture() or nil
            sliceText = recipe:getName()
        else
            icon = nil
        end
        menu:addSlice(sliceText, icon, self.onSliceSelected, self, i)
    end

    self:display()
end

function ISCustomRadialMenu:configEmote()
	local ui = ISUIEmoteConfig:new(0, 0, self.character);
	ui:initialise();
	ui:addToUIManager();
end

----------------------------------

local STATE = {}
STATE[1] = {}
STATE[2] = {}
STATE[3] = {}
STATE[4] = {}

function ISCustomRadialMenu.checkKey(key)
	return true
end

function ISCustomRadialMenu.onKeyPressed(key)
	if not ISCustomRadialMenu.checkKey(key) then
		return
	end
	if getCore():getGameMode() == "Tutorial" and not getCore():isKey("Shout", key) then
		return;
	end
	local radialMenu = getPlayerRadialMenu(0)
	if getCore():getOptionRadialMenuKeyToggle() and radialMenu:isReallyVisible() then
		STATE[1].radialWasVisible = true
		radialMenu:removeFromUIManager()
		return
	end
	STATE[1].keyPressedMS = getTimestampMs()
	STATE[1].radialWasVisible = false
end

function ISCustomRadialMenu.onKeyRepeat(key)
	if not ISCustomRadialMenu.checkKey(key) then
		return
	end
end

function ISCustomRadialMenu.onKeyReleased(key)
	if not ISCustomRadialMenu.checkKey(key) then
		return
	end
end

local function OnGameStart()
	Events.OnKeyStartPressed.Add(ISCustomRadialMenu.onKeyPressed)
	Events.OnKeyKeepPressed.Add(ISCustomRadialMenu.onKeyRepeat)
	Events.OnKeyPressed.Add(ISCustomRadialMenu.onKeyReleased)
end

Events.OnGameStart.Add(OnGameStart)
